#include <filesystem>
#include <fstream>

#include <opencv2/opencv.hpp>
#include <botan/base64.h>

#include "controllers/controllers.h"
#include "dto/dto.h"

// https://developer.mozilla.org/en-US/docs/Web/Media/Formats/Image_types#common_image_file_types
const std::unordered_map<std::string, std::string> mime_type_map = {
        { ".apng" , "image/apng" },
        { ".avif" , "image/avif" },
        { ".bmp"  , "image/bmp"  },
        { ".gif"  , "image/gif"  },
        { ".jpg"  , "image/jpeg" },
        { ".jpeg" , "image/jpeg" },
        { ".jfif" , "image/jpeg" },
        { ".pjpeg", "image/jpeg" },
        { ".pjp"  , "image/jpeg" },
        { ".png"  , "image/png"  },
        { ".svg"  , "image/svg"  },
        { ".webp" , "image/webp" },

        { ".aac"  , "audio/aac"  },
        { ".flac" , "audio/flac" },
        { ".mp3"  , "audio/mp3"  },
        { ".m4a"  , "audio/mp4"  },
        { ".oga"  , "audio/ogg"  },
        { ".ogg"  , "audio/ogg"  },
        { ".wav"  , "audio/wav"  },

        { ".3gp"  , "video/3gpp" },
        { ".mpg"  , "video/mpeg" },
        { ".mpeg" , "video/mpeg" },
        { ".mp4"  , "video/mp4"  },
        { ".m4v"  , "video/mp4"  },
        { ".m4p"  , "video/mp4"  },
        { ".ogv"  , "video/ogg"  },
        { ".mov"  , "video/quicktime" },
        { ".webm" , "video/webm" },
        { ".mkv"  , "video/x-matroska" },
        { ".mk3d" , "video/x-matroska" },
        { ".mks"  , "video/x-matroska" },

        { ".pdf"  , "application/pdf" }
};

template<typename InputIt>
std::string join_string(InputIt first, InputIt last, const std::string& separator = ",") {
    std::ostringstream result;
    if (first != last) {
        result << *first;
        while (++first != last) {
            result << separator << *first;
        }
    }
    return result.str();
}

namespace api {
    void fs::root(req_type req, cbk_type cbk) {
        db::User user = dto::get_user(req);
        cbk(dto::Responses::get_root_res(user.getValueOfRootId()));
    }

    void fs::node(req_type req, cbk_type cbk, uint64_t node) {
        db::User user = dto::get_user(req);
        std::shared_lock lock(*get_user_mutex(user.getValueOfId()));
        auto inode =  get_node_and_validate(user, node);
        if (!inode.has_value())
            return cbk(dto::Responses::get_badreq_res("Unknown node"));
        auto dto_node = dto::Responses::GetNodeEntry(*inode);
        std::vector<dto::Responses::GetNodeEntry> children;
        if (!dto_node.is_file) for (const db::INode& child : get_children(*inode)) children.emplace_back(child);
        cbk(dto::Responses::get_node_res(dto_node, children));
    }

    void fs::path(req_type req, cbk_type cbk, uint64_t node) {
        db::User user = dto::get_user(req);
        std::shared_lock lock(*get_user_mutex(user.getValueOfId()));
        auto inode = get_node_and_validate(user, node);
        if (!inode.has_value())
            cbk(dto::Responses::get_badreq_res("Unknown node"));
        else {
            auto path = generate_path(*inode);
            cbk(dto::Responses::get_success_res(path));
        }
    }

    template<bool file>
    void fs::create_node_req(req_type req, cbk_type cbk) {
        db::User user = dto::get_user(req);
        Json::Value& json = *req->jsonObject();
        try {
            uint64_t parent = dto::json_get<uint64_t>(json, "parent").value();
            std::string name = dto::json_get<std::string>(json, "name").value();

            std::shared_lock lock(*get_user_mutex(user.getValueOfId()));

            auto new_node = create_node(name, user, file, std::make_optional(parent));
            if (std::holds_alternative<db::INode>(new_node))
                cbk(dto::Responses::get_new_node_res(std::get<db::INode>(new_node).getValueOfId()));
            else if (std::holds_alternative<create_node_error>(new_node))
                switch (std::get<create_node_error>(new_node)) {
                    case create_node_error::INVALID_NAME: return cbk(dto::Responses::get_badreq_res("Invalid name"));
                    case create_node_error::INVALID_PARENT: return cbk(dto::Responses::get_badreq_res("Invalid parent"));
                    case create_node_error::FILE_PARENT: return cbk(dto::Responses::get_badreq_res("Parent is file"));
                }
            else {
                auto tuple = std::get<std::tuple<bool, uint64_t>>(new_node);
                cbk(dto::Responses::get_node_exists_res(std::get<1>(tuple), std::get<0>(tuple)));
            }
        } catch (const std::exception&) {
            cbk(dto::Responses::get_badreq_res("Validation error"));
        }
    }

    void fs::delete_node_req(req_type req, cbk_type cbk, uint64_t node) {
        db::User user = dto::get_user(req);
        std::unique_lock lock(*get_user_mutex(user.getValueOfId()));
        auto inode = get_node_and_validate(user, node);
        if (!inode.has_value())
            cbk(dto::Responses::get_badreq_res("Unknown node"));
        else if (inode->getValueOfParentId() == 0)
            cbk(dto::Responses::get_badreq_res("Can't delete root"));
        else {
            auto chan = std::make_shared<msd::channel<std::string>>();
            std::string("Waiting in queue...\n") >> (*chan);
            get_delete_loop()->queueInLoop([chan, inode=*inode, user=user.getValueOfId()]{
                std::unique_lock lock(*get_user_mutex(user));
                delete_node(inode, *chan);
                chan->close();
            });
            cbk(drogon::HttpResponse::newStreamResponse([chan](char* buf, std::size_t size) -> std::size_t{
                if (buf == nullptr) return 0;
                if (chan->closed() && chan->empty()) return 0;
                std::string buffer;
                buffer << *chan;
                if (buffer.empty()) return 0;
                std::size_t read = std::min(size, buffer.size());
                std::memcpy(buf, buffer.data(), read); // NOLINT(bugprone-not-null-terminated-result)
                return read;
            }));
        }
    }

    void fs::upload(req_type req, cbk_type cbk, uint64_t node) {
        constexpr int image_height = 256;
        db::User user = dto::get_user(req);

        std::shared_lock lock(*get_user_mutex(user.getValueOfId()));

        auto inode = get_node_and_validate(user, node);
        if (!inode.has_value())
            return cbk(dto::Responses::get_badreq_res("Unknown node"));
        if (inode->getValueOfIsFile() == 0)
            return cbk(dto::Responses::get_badreq_res("Can't upload to a directory"));

        drogon::MultiPartParser mpp;
        if (mpp.parse(req) != 0)
            return cbk(dto::Responses::get_badreq_res("Failed to parse files"));
        if (mpp.getFiles().size() != 1)
            return cbk(dto::Responses::get_badreq_res("Exactly 1 file needed"));

        const drogon::HttpFile& file = mpp.getFiles().at(0);

        std::filesystem::path p("./files");
        p /= std::to_string(inode->getValueOfId());

        file.saveAs(p.string());
        try {
            if (file.fileLength() > 100 * 1024 * 1024) throw std::exception();
            std::filesystem::path filename(inode->getValueOfName());
            std::string ext = filename.extension().string();
            std::transform(ext.begin(), ext.end(), ext.begin(), tolower);
            const std::string& mime = mime_type_map.at(ext);
            if (!mime.starts_with("image")) throw std::exception();

            cv::_InputArray image_arr(file.fileData(), (int) file.fileLength());
            cv::Mat image = cv::imdecode(image_arr, cv::IMREAD_COLOR);
            if (!image.empty()) {
                float h_ration = ((float) image_height) / ((float) image.rows);
                cv::Mat preview;
                cv::resize(image, preview, cv::Size((int) (((float) image.cols) * h_ration), image_height), 0, 0, cv::INTER_AREA);
                cv::imwrite(p.string() + "_preview.png", preview);
                inode->setHasPreview(1);
            }
        } catch (const std::exception&) {}
        inode->setSize(file.fileLength());

        db::MapperInode inode_mapper(drogon::app().getDbClient());
        inode_mapper.update(*inode);

        cbk(dto::Responses::get_success_res());
    }

    void fs::create_zip(req_type req, cbk_type cbk) {
        db::User user = dto::get_user(req);

        std::shared_lock lock(*get_user_mutex(user.getValueOfId()));

        Json::Value& json = *req->jsonObject();
        try {
            if (!json.isMember("nodes")) throw std::exception();
            Json::Value node_arr = json["nodes"];
            if (!node_arr.isArray()) throw std::exception();
            std::vector<uint64_t> node_ids;
            for (const auto& node : node_arr)
                node_ids.push_back(node.asUInt64());

            std::vector<db::INode> nodes;
            std::transform(node_ids.begin(), node_ids.end(), std::back_inserter(nodes), [&user](uint64_t node) {
                return api::fs::get_node_and_validate(user, node).value();
            });

            std::string key = join_string(node_ids.begin(), node_ids.end());

            if (zip_to_temp_map.contains(key)) return cbk(dto::Responses::get_create_zip_done_res());
            if (in_progress_zips.contains(key)) {
                auto progress = in_progress_zips.at(key);
                return cbk(dto::Responses::get_create_zip_done_res(std::get<1>(progress), std::get<2>(progress)));
            }
            std::string file_name = "./temp/fs_" + std::to_string(next_temp_id++) + ".zip";
            in_progress_zips.emplace(key, std::make_tuple(file_name, 0, 1));
            get_zip_loop()->queueInLoop([key = std::move(key), nodes = std::move(nodes), file_name = std::move(file_name), user=user.getValueOfId()]{
                {
                    std::shared_lock lock(*get_user_mutex(user));
                    uint64_t size = 0;
                    for (const auto& node : nodes) size += calc_total_size(node);
                    std::get<2>(in_progress_zips.at(key)) = size;
                    struct zip_t* zip = zip_open(file_name.c_str(), ZIP_DEFAULT_COMPRESSION_LEVEL, 'w');
                    for (const db::INode& node : nodes)
                        add_to_zip(zip, key, node, "");
                    zip_close(zip);
                }
                zip_to_temp_map.emplace(key, file_name);
                in_progress_zips.erase(key);
            });
            return cbk(dto::Responses::get_create_zip_done_res(0, 1));
        } catch (const std::exception&) {
            cbk(dto::Responses::get_badreq_res("Validation error"));
        }
    }

    void fs::download(req_type req, cbk_type cbk) {
        db::User user = dto::get_user(req);

        std::shared_lock lock(*get_user_mutex(user.getValueOfId()));

        auto node_id = req->getOptionalParameter<uint64_t>("id");
        if (!node_id.has_value()) {
            cbk(dto::Responses::get_badreq_res("Invalid node"));
            return;
        }
        auto inode = get_node_and_validate(user, *node_id);
        if (!inode.has_value()) {
            cbk(dto::Responses::get_badreq_res("Invalid node"));
            return;
        }

        if (inode->getValueOfIsFile() != 0) {
            std::filesystem::path p("./files");
            p /= std::to_string(inode->getValueOfId());

            cbk(drogon::HttpResponse::newFileResponse(
                    p.string(),
                    inode->getValueOfName()
            ));
        } else {
            try {
                std::string key = std::to_string(inode->getValueOfId());
                std::string file = zip_to_temp_map.at(key);
                zip_to_temp_map.erase(key);
                cbk(drogon::HttpResponse::newFileResponse(
                        file,
                        inode->getValueOfName() + ".zip"
                ));
            } catch (const std::exception&) {
                cbk(dto::Responses::get_badreq_res("Invalid node"));
            }
        }
    }

    void fs::download_multi(req_type req, cbk_type cbk) {
        db::User user = dto::get_user(req);

        std::shared_lock lock(*get_user_mutex(user.getValueOfId()));

        auto node_ids_str = req->getOptionalParameter<std::string>("id");
        if (!node_ids_str.has_value())
            return cbk(dto::Responses::get_badreq_res("No nodes"));

        std::stringstream node_ids_ss(*node_ids_str);
        std::string temp;
        try {
            while (std::getline(node_ids_ss, temp, ','))
                if (!get_node_and_validate(user, std::stoull(temp)).has_value()) throw std::exception();

            std::string file = zip_to_temp_map.at(*node_ids_str);
            zip_to_temp_map.erase(*node_ids_str);
            cbk(drogon::HttpResponse::newFileResponse(
                    file,
                    "files.zip"
            ));
        } catch (const std::exception&) {
            cbk(dto::Responses::get_badreq_res("Invalid nodes"));
        }
    }

    void fs::download_preview(req_type req, cbk_type cbk, uint64_t node) {
        db::User user = dto::get_user(req);

        std::shared_lock lock(*get_user_mutex(user.getValueOfId()));

        auto inode = get_node_and_validate(user, node);
        if (!inode.has_value())
            return cbk(dto::Responses::get_badreq_res("Unknown node"));
        if (inode->getValueOfHasPreview() == 0)
            return cbk(dto::Responses::get_badreq_res("No preview"));

        std::filesystem::path p("./files");
        p /= std::to_string(inode->getValueOfId()) + "_preview.png";
        std::ifstream file(p, std::ios::in | std::ios::binary);
        std::vector<uint8_t> image((std::istreambuf_iterator<char>(file)), std::istreambuf_iterator<char>());

        cbk(dto::Responses::get_download_base64_res("data:image/png;base64," + Botan::base64_encode(image)));
    }

    void fs::get_type(req_type req, cbk_type cbk, uint64_t node){
        db::User user = dto::get_user(req);

        std::shared_lock lock(*get_user_mutex(user.getValueOfId()));

        auto inode = get_node_and_validate(user, node);
        if (!inode.has_value())
            return cbk(dto::Responses::get_badreq_res("Unknown node"));

        std::filesystem::path name(inode->getValueOfName());
        std::string ext = name.extension().string();
        std::transform(ext.begin(), ext.end(), ext.begin(), tolower);

        try {
            cbk(dto::Responses::get_type_res(mime_type_map.at(ext)));
        } catch (const std::exception&) {
            cbk(dto::Responses::get_badreq_res("Invalid file type"));
        }
    }
}